program main
  ! This program fits a quadratic function using a small neural network using
  ! stochastic gradient descent, batch gradient descent, and minibatch gradient
  ! descent.
  use nf, only: dense, input, network
  use nf_dense_layer, only: dense_layer
  use nf_optimizers, only: sgd, rmsprop, adam, adagrad

  implicit none
  type(network) :: net(11)

  ! Training parameters
  integer, parameter :: num_epochs    = 1000
  
  integer, parameter :: train_size    = 7000
  integer, parameter :: test_size     = 100
  
  integer, parameter :: batch_size    = 100
  real   , parameter :: learning_rate = 0.001
  real   , parameter :: decay_rate    = 0.9
  real   , parameter :: beta1         = 0.85
  real   , parameter :: beta2         = 0.95
  real   , parameter :: epsilon       = 1e-8
  character(Len=528) :: path_in,path_out
  character(Len=528) :: filename_train,filename_test

  ! Input and output data
  real, allocatable :: xr(:),x(:), y(:)       ! training data
  real, allocatable :: xtest(:), ytest(:) ! testing data

  integer :: i, n

  print '("Fitting quadratic function")'
  print '(60("="))'
  ! xtest and ytest as 1-D arrays
  allocate(xtest(test_size), ytest(test_size))
  ! x and y as 1-D arrays
  allocate(x(train_size),xr(train_size), y(train_size))

  path_in='/cygdrive/d/paulo.kubota/pgmet/isabella/neural-fortran/example/isabella/datain'
  path_out='/cygdrive/d/paulo.kubota/pgmet/isabella/neural-fortran/example/isabella/datain'
  filename_train='datatrain.dat'
  filename_test='datatest.dat'

  OPEN(10,file=TRIM(path_in)//'/'//trim(filename_train),status='old',action='read',form='formatted') 
  read(10,'(1X)')
  do i=1, train_size
    read (10,*)x(i),y(i)
  end do

  OPEN(20,file=TRIM(path_in)//'/'//trim(filename_test),status='old',action='read',form='formatted') 
  read(20,'(1X)')
  do i=1, test_size
    read (20,*)xtest(i),ytest(i)
  end do

  print '("Simple")'
  print '(60("="))'  
  !xtest = [(real(i - 1) * 2 / test_size, i = 1, test_size)]
  !ytest = gaussian_fit(xtest)


  ! Generating the dataset
  do i = 1, train_size
    xr(i) = x(i)!ABS(x(i)/MAXVAL(ABS(x)))
    y(i)  = gaussian_fit(xr(i))
  end do

  ! Instantiate a network and copy an instance to the rest of the array
  net(1) = network([input(1), dense(2),dense(3), dense(2),dense(1)])
  !net(2:) = net(1)

  ! Print network info to stdout; this will be the same for all three networks.
  call net(1) % print_info()

  ! SGD, no momentum
  call sgd_optimizer(net(1), x, y, xtest, ytest, learning_rate, num_epochs)


contains

  real elemental function gaussian_fit(x) result(y)
    ! Quadratic function
    real, intent(in) :: x
    !y = (x**2 / 2 + x / 2 + 1) / 2
    y = 0.002 + 3.2e-2 * exp (- (( x - 0.0)**2)/(2*(0.007696)**2))
  end function gaussian_fit

  subroutine sgd_optimizer( &
    net, x, y, xtest, ytest, learning_rate, num_epochs, momentum, nesterov &
  )
    ! In the stochastic gradient descent (SGD) optimizer, we run the forward
    ! and backward passes and update the weights for each training sample,
    ! one at a time.
    type(network), intent(inout) :: net
    real, intent(in) :: x(:), y(:)
    real :: xx(1), yy(1)
    real, intent(in) :: xtest(:), ytest(:)
    real, intent(in) :: learning_rate
    integer, intent(in) :: num_epochs
    real, intent(in), optional :: momentum
    logical, intent(in), optional :: nesterov
    real :: ypred(size(xtest))
    real :: momentum_value
    logical :: nesterov_value
    integer :: i, n

    print '(a)', 'Stochastic gradient descent'
    print '(34("-"))'

    ! Set default values for momentum and nesterov
    if (.not. present(momentum)) then
      momentum_value = 0.0
    else
      momentum_value = momentum
    end if

    if (.not. present(nesterov)) then
      nesterov_value = .false.
    else
      nesterov_value = nesterov
    end if

    do n = 1, num_epochs
      do i = 1, size(x)
        call net % forward([x(i)])
        call net % backward([y(i)])
        call net % update( &
          sgd( &
            learning_rate=learning_rate, &
            momentum=momentum_value, &
            nesterov=nesterov_value &
          ) &
        )
      end do

      if (mod(n, num_epochs / 10) == 0) then
        DO i = 1, size(xtest)
           xx=xtest(i)
           yy= net % predict(xx)
           ypred(i)=yy(1)
           print*,xtest(i),ypred(i),ytest(i),ypred(i)-ytest(i)
        END DO 
        print '("Epoch: ", i4,"/",i4,", RMSE = ", f9.6)', &
             n, num_epochs, sum((ypred - ytest)**2) / size(ytest)
      end if

    end do

    print *, ''

  end subroutine sgd_optimizer

  subroutine shuffle(arr)
    ! Shuffle an array using the Fisher-Yates algorithm.
    integer, intent(inout) :: arr(:)
    real :: a
    integer :: i, temp

    do i = size(arr), 2, -1
      call random_number(a)
      a = floor(a * real(i)) + 1
      temp = arr(i)
      arr(i) = arr(int(a))
      arr(int(a)) = temp
    end do
  end subroutine shuffle

end program main
